/* Fesslix - Stochastic Analysis
 * Copyright (C) 2010-2022 Wolfgang Betz
 *
 * Fesslix is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Fesslix is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Fesslix.  If not, see <http://www.gnu.org/licenses/>. 
 */

#define FLXLIB_CPP

#include "flxostools.h"

#include <stdio.h>  /* defines FILENAME_MAX */
#ifdef __WINDOWS__
  #include <direct.h>
  #define GetCurrentDir _getcwd
#else
  #include <unistd.h>
  #define GetCurrentDir getcwd
#endif

#ifdef __unix__
  #include <pwd.h>
#elif defined(__WINDOWS__)
  #include <windows.h>
  #include "Userenv.h"
  #pragma comment(lib, "userenv.lib")
#else
  #error Operating system not handled.
#endif

#ifdef __unix__
  #include <termios.h>
#elif defined(__WINDOWS__)
  #include <stdlib.h>
#else
  #error Operating system not handled.
#endif

#include <limits.h>


char cCurrentPath[FILENAME_MAX];

const bool get_config_dir_global(std::string& configDir)
{
  #ifdef __unix__
    configDir = "/etc/fesslix/";
  #elif defined(__WINDOWS__)
    configDir = GlobalVar.get_exe_dir() + '/';
  #endif
  return true;
}

const bool get_config_dir_local(std::string& configDir)
{
  #ifdef __unix__
    configDir = getpwuid(getuid())->pw_dir;
    configDir+="/.config/fesslix/";
  #elif defined(__WINDOWS__)
    configDir = "";
  #endif
  return true;
}

const bool get_pwd(std::string& pwdStr) 
{
  if (!GetCurrentDir(cCurrentPath, sizeof(cCurrentPath) / sizeof(char)))
  {
    return false;
  }
  cCurrentPath[sizeof(cCurrentPath) - 1] = '\0'; /* not really required */
  pwdStr = cCurrentPath;
  return true;
}

void setEXEdir()
{
  std::string resstr;
  #ifdef __WINDOWS__
    char exePath[MAX_PATH]; 
    HMODULE hModule = GetModuleHandle(NULL);
    if (hModule) {
     GetModuleFileName(hModule,exePath, (sizeof(exePath))); 
     resstr = exePath;
    } else {
     throw FlxException("getEXEdir_1","Module handle is NULL");
    }
  #else
    char buff[PATH_MAX];
    ssize_t len = ::readlink("/proc/self/exe", buff, sizeof(buff)-1);
    if (len != -1) {
      buff[len] = '\0';
      resstr = std::string(buff);
    } else {
      throw FlxException("getEXEdir_2","readlink returned an error");
    }
  #endif
  if (resstr.empty()) throw FlxException_Crude("getEXEdir_3");
  GlobalVar.set_exe_PathName(resstr);
  resstr = resstr.substr(0, resstr.find_last_of("/\\")+1);
  GlobalVar.set_exe_dir(resstr);
}

#ifdef __unix__
int mygetch ( void ) 
{
  int ch;
  struct termios oldt, newt;
  
  tcgetattr ( STDIN_FILENO, &oldt );
  newt = oldt;
  newt.c_lflag &= ~( ICANON | ECHO );
  tcsetattr ( STDIN_FILENO, TCSANOW, &newt );
  ch = getchar();
  tcsetattr ( STDIN_FILENO, TCSANOW, &oldt );
  
  return ch;
}
#endif

void flx_take_a_break()
{
  #ifdef __WINDOWS__
    system ( "PAUSE" );
  #elif defined(__unix__)
    *(GlobalVar.get_cout()) <<"Press any key to continue . . ." ;
    mygetch();
    *(GlobalVar.get_cout()) << std::endl;
  #else
    *(GlobalVar.get_cout()) <<"Press [Enter] to continue . . .";
    std::cin.get();
  #endif
}

void flx_put_to_sleep(const tuint seconds)
{
  #ifdef __WINDOWS__
    Sleep(seconds*1000);
  #else
    sleep(seconds);
  #endif
}


